@@ -12,21 +12,17 @@ from future.utils import viewitems, viewvalues
 
 import json
 import logging
-import random
 from copy import deepcopy
 from rucio.client import Client
 from rucio.common.exception import (AccountNotFound, DataIdentifierNotFound, AccessDenied, DuplicateRule,
                                     DataIdentifierAlreadyExists, DuplicateContent, InvalidRSEExpression,
-                                    UnsupportedOperation, FileAlreadyExists, RuleNotFound, RSENotFound)
+                                    UnsupportedOperation, FileAlreadyExists, RuleNotFound, RSENotFound, RuleReplaceFailed)
 from Utils.MemoryCache import MemoryCache
+from Utils.IteratorTools import grouper
+from WMCore.Services.Rucio.RucioUtils import (validateMetaData, weightedChoice,
+                                              isTapeRSE, dropTapeRSEs)
 from WMCore.WMException import WMException
 
-RUCIO_VALID_PROJECT = ("Production", "RelVal", "Tier0", "Test", "User")
-# grouping values are extracted from:
-# https://github.com/rucio/rucio/blob/master/lib/rucio/common/schema/cms.py#L117
-GROUPING_DSET = "DATASET"
-GROUPING_ALL = "ALL"
-
 
 class WMRucioException(WMException):
     """
@@ -44,64 +40,6 @@ class WMRucioDIDNotFoundException(WMException):
     pass
 
 
-def validateMetaData(did, metaDict, logger):
-    """
-    This function can be extended in the future, for now it will only
-    validate the DID creation metadata, more specifically only the
-    "project" parameter
-    :param did: the DID that will be inserted
-    :param metaDict: a dictionary with all the DID metadata data to be inserted
-    :param logger: a logger object
-    :return: False if validation fails, otherwise True
-    """
-    if metaDict.get("project", "Production") in RUCIO_VALID_PROJECT:
-        return True
-    msg = "DID: %s has an invalid 'project' meta-data value: %s" % (did, metaDict['project'])
-    msg += "The supported 'project' values are: %s" % str(RUCIO_VALID_PROJECT)
-    logger.error(msg)
-    return False
-
-
-def weightedChoice(choices):
-    # from https://stackoverflow.com/questions/3679694/a-weighted-version-of-random-choice
-    # Python 3.6 includes something like this in the random library itself
-
-    total = sum(w for c, w in choices)
-    r = random.uniform(0, total)
-    upto = 0
-    for c, w in choices:
-        if upto + w >= r:
-            return c
-        upto += w
-    assert False, "Shouldn't get here"
-
-
-def isTapeRSE(rseName):
-    """
-    Given an RSE name, return True if it's a Tape RSE (rse_type=TAPE), otherwise False
-    :param rseName: string with the RSE name
-    :return: True or False
-    """
-    # NOTE: a more reliable - but more expensive - way to know that would be
-    # to query `get_rse` and evaluate the rse_type parameter
-    return rseName.endswith("_Tape")
-
-
-def dropTapeRSEs(listRSEs):
-    """
-    Method to parse a list of RSE names and return only those that
-    are not a rse_type=TAPE, so in general only Disk endpoints
-    :param listRSEs: list with the RSE names
-    :return: a new list with only DISK RSE names
-    """
-    diskRSEs = []
-    for rse in listRSEs:
-        if rse.endswith("_Tape"):
-            continue
-        diskRSEs.append(rse)
-    return diskRSEs
-
-
 class Rucio(object):
     """
     Service class providing additional Rucio functionality on top of the
@@ -115,7 +53,7 @@ class Rucio(object):
     We will try to use Container -> CMS dataset, block -> CMS block.
     """
 
-    def __init__(self, acct, hostUrl=None, authUrl=None, configDict=None):
+    def __init__(self, acct, hostUrl=None, authUrl=None, configDict=None, client=None):
         """
         Constructs a Rucio object with the Client object embedded.
         In order to instantiate a Rucio Client object, it assumes the host has
@@ -125,6 +63,7 @@ class Rucio(object):
         :param hostUrl: defaults to the rucio config one
         :param authUrl: defaults to the rucio config one
         :param configDict: dictionary with extra parameters
+        :param client: optional Rucio client to use (useful for mock-up)
         """
         configDict = configDict or {}
         # default RSE data caching to 12h
@@ -142,15 +81,18 @@ class Rucio(object):
         self.rucioParams.setdefault('user_agent', 'wmcore-client')
 
         self.logger.info("WMCore Rucio initialization parameters: %s", self.rucioParams)
-        self.cli = Client(rucio_host=hostUrl, auth_host=authUrl, account=acct,
-                          ca_cert=self.rucioParams['ca_cert'], auth_type=self.rucioParams['auth_type'],
-                          creds=self.rucioParams['creds'], timeout=self.rucioParams['timeout'],
-                          user_agent=self.rucioParams['user_agent'])
-        clientParams = {}
-        for k in ("host", "auth_host", "auth_type", "account", "user_agent",
-                  "ca_cert", "creds", "timeout", "request_retries"):
-            clientParams[k] = getattr(self.cli, k)
-        self.logger.info("Rucio client initialization parameters: %s", clientParams)
+        if client:
+            self.cli = client
+        else:
+            self.cli = Client(rucio_host=hostUrl, auth_host=authUrl, account=acct,
+                              ca_cert=self.rucioParams['ca_cert'], auth_type=self.rucioParams['auth_type'],
+                              creds=self.rucioParams['creds'], timeout=self.rucioParams['timeout'],
+                              user_agent=self.rucioParams['user_agent'])
+            clientParams = {}
+            for k in ("host", "auth_host", "auth_type", "account", "user_agent",
+                      "ca_cert", "creds", "timeout", "request_retries"):
+                clientParams[k] = getattr(self.cli, k)
+            self.logger.info("Rucio client initialization parameters: %s", clientParams)
 
         # keep a map of rse expression to RSE names mapped for some time
         self.cachedRSEs = MemoryCache(rseCacheExpiration, {})
@@ -211,7 +153,7 @@ class Rucio(object):
         :return: a list of dictionaries with the account usage information.
           None in case of failure
         """
-        res = None
+        res = []
         try:
             res = list(self.cli.get_local_account_usage(acct, rse=rse))
         except (AccountNotFound, AccessDenied) as ex:
@@ -228,7 +170,7 @@ class Rucio(object):
         :return: a list of block names
         """
         blockNames = []
-        if not self.isContainer(container):
+        if not self.isContainer(container, scope=scope):
             # input container wasn't really a container
             self.logger.warning("Provided DID name is not a CONTAINER type: %s", container)
             return blockNames
@@ -280,7 +222,7 @@ class Rucio(object):
                 for item in self.cli.list_dataset_replicas(kwargs["scope"], did["name"], deep=kwargs['deep']):
                     resultDict.setdefault(item['name'], [])
                     if item['state'].upper() == 'AVAILABLE':
-                        resultDict[item['name']].append(item['rse'])            
+                        resultDict[item['name']].append(item['rse'])
         else:
             for item in self.cli.list_dataset_replicas_bulk(inputDids):
                 resultDict.setdefault(item['name'], [])
@@ -303,7 +245,7 @@ class Rucio(object):
         See here for documentation of the upstream Rucio API: https://rucio.readthedocs.io/en/latest/api/rse.html
         :param site: a Rucio RSE, i.e. a site name in standard CMS format like 'T1_UK_RAL_Disk' or  'T2_CH_CERN'
         :param lfns: one LFN or a list of LFN's, does not need to correspond to actual files and could be a top level directory
-                      like ['/store/user/rucio/jdoe','/store/data',...] or the simple '/store/data' 
+                      like ['/store/user/rucio/jdoe','/store/data',...] or the simple '/store/data'
         :param protocol: If the RSE supports multiple access protocols, a preferred protocol can be selected via this,
                          otherwise the default one for the site will be selected. Example: 'gsiftp' or 'davs'
                          this is what is called "scheme" in the RUCIO API (we think protocol is more clear)
@@ -405,7 +347,7 @@ class Rucio(object):
             response = self.attachDIDs(kwargs.get('rse'), container, name, scope)
         return response
 
-    def attachDIDs(self, rse, superDID, dids, scope='cms'):
+    def attachDIDs(self, rse, superDID, dids, scope='cms', chunkSize=1000):
         """
         _attachDIDs_
 
@@ -415,26 +357,38 @@ class Rucio(object):
              then it's a container name; if attaching files, then it's a block name)
         :param dids: either a string or a list of data identifiers (can be block or files)
         :param scope: string with the scope name
+        :param chunkSize: maximum number of dids to be attached in any single Rucio server call
         :return: a boolean to represent whether it succeeded or not
         """
         if not isinstance(dids, list):
             dids = [dids]
-        dids = [{'scope': scope, 'name': did} for did in dids]
+        # NOTE: the attaching dids do not create new container within a scope
+        # and it is safe to use cms scope for it
+        alldids = [{'scope': 'cms', 'name': did} for did in sorted(dids)]
+
+        # report if we use chunk size in rucio attach_dids API call
+        if len(alldids) > chunkSize:
+            self.logger.info("Attaching a total of %d DIDs in chunk size of: %d", len(alldids), chunkSize)
+
+        for dids in grouper(alldids, chunkSize):
+            response = False
+            try:
+                response = self.cli.attach_dids(scope, superDID, dids=dids, rse=rse)
+            except DuplicateContent:
+                self.logger.warning("Dids: %s already attached to: %s", dids, superDID)
+                response = True
+            except FileAlreadyExists:
+                self.logger.warning("Failed to attach files already existent on block: %s", superDID)
+                response = True
+            except DataIdentifierNotFound:
+                self.logger.error("Failed to attach dids: %s. Parent DID %s does not exist.", dids, superDID)
+            except Exception as ex:
+                self.logger.error("Exception attaching %s dids to: %s. Error: %s. First 10 dids: %s",
+                                  len(dids), superDID, str(ex), dids[:10])
+            if not response:
+                # if we had failure with specific chunk of dids we'll return immediately
+                return response
 
-        response = False
-        try:
-            response = self.cli.attach_dids(scope, superDID, dids=dids, rse=rse)
-        except DuplicateContent:
-            self.logger.warning("Dids: %s already attached to: %s", dids, superDID)
-            response = True
-        except FileAlreadyExists:
-            self.logger.warning("Failed to attach files already existent on block: %s", superDID)
-            response = True
-        except DataIdentifierNotFound:
-            self.logger.error("Failed to attach dids: %s. Parent DID %s does not exist.", dids, superDID)
-        except Exception as ex:
-            self.logger.error("Exception attaching dids: %s to: %s. Error: %s",
-                              dids, superDID, str(ex))
         return response
 
     def createReplicas(self, rse, files, block, scope='cms', ignoreAvailability=True):
@@ -500,6 +454,30 @@ class Rucio(object):
             self.logger.error("Exception closing container/block: %s. Error: %s", name, str(ex))
         return response
 
+    def moveReplicationRule(self, ruleId, rseExpression, account):
+        """
+        Perform move operation for provided rule id and rse expression
+        :param ruleId: rule id
+        :param rseExpression: rse expression
+        :param account: rucio quota account
+        :return: it returns either an empty list or a list with a string id for the rule created
+        Please note, we made return type from this wrapper compatible with createReplicateRule
+        """
+        ruleIds = []
+        try:
+            rid = self.cli.move_replication_rule(ruleId, rseExpression, account)
+            ruleIds.append(rid)
+        except RuleNotFound as ex:
+            msg = "RuleNotFound move DID replication rule. Error: %s" % str(ex)
+            raise WMRucioException(msg) from ex
+        except RuleReplaceFailed as ex:
+            msg = "RuleReplaceFailed move DID replication rule. Error: %s" % str(ex)
+            raise WMRucioException(msg) from ex
+        except Exception as ex:
+            msg = "Unsupported exception from Rucio API. Error: %s" % str(ex)
+            raise WMRucioException(msg) from ex
+        return ruleIds
+
     def createReplicationRule(self, names, rseExpression, scope='cms', copies=1, **kwargs):
         """
         _createReplicationRule_
@@ -705,24 +683,39 @@ class Rucio(object):
             self.logger.error("Exception listing parent DIDs for data: %s. Error: %s", name, str(ex))
         return list(res)
 
-    def getRule(self, ruleId, estimatedTtc=False):
+    def getRule(self, ruleId):
         """
         _getRule_
 
         Retrieve rule information for a given rule id
         :param ruleId: string with the rule id
-        :param estimatedTtc: bool, if rule_info should return ttc information
         :return: a dictionary with the rule data (or empty if no rule can be found)
         """
         res = {}
         try:
-            res = self.cli.get_replication_rule(ruleId, estimate_ttc=estimatedTtc)
+            res = self.cli.get_replication_rule(ruleId)
         except RuleNotFound:
             self.logger.error("Cannot find any information for rule id: %s", ruleId)
         except Exception as ex:
             self.logger.error("Exception getting rule id: %s. Error: %s", ruleId, str(ex))
         return res
 
+    def updateRule(self, ruleId, opts):
+        """
+        Update rule information for a given rule id
+        :param ruleId: string with the rule id
+        :param opts: dictionary, rule id options passed to Rucio
+        :return: boolean status of update call
+        """
+        status = None
+        try:
+            status = self.cli.update_replication_rule(ruleId, opts)
+        except RuleNotFound:
+            self.logger.error("Cannot find any information for rule id: %s", ruleId)
+        except Exception as ex:
+            self.logger.error("Exception updating rule id: %s. Error: %s", ruleId, str(ex))
+        return status
+
     def deleteRule(self, ruleId, purgeReplicas=False):
         """
         _deleteRule_
@@ -764,8 +757,9 @@ class Rucio(object):
             except InvalidRSEExpression as exc:
                 msg = "Provided RSE expression is considered invalid: {}. Error: {}".format(rseExpr, str(exc))
                 raise WMRucioException(msg)
-        # add this key/value pair to the cache
-        self.cachedRSEs.addItemToCache({rseExpr: matchingRSEs})
+        if useCache:
+            # add this key/value pair to the cache
+            self.cachedRSEs.addItemToCache({rseExpr: matchingRSEs})
         if returnTape:
             return matchingRSEs
         return dropTapeRSEs(matchingRSEs)
@@ -774,7 +768,7 @@ class Rucio(object):
         """
         _pickRSE_
 
-        Use a weighted random selection algorithm to pick an RSE for a dataset based on an attribute
+        Use a weighted random selection algorithm to pick an RSE for a dataset based on an RSE attribute.
         The attribute should correlate to space available.
         :param rseExpression: Rucio RSE expression to pick RSEs (defaults to production Tape RSEs)
         :param rseAttribute: The RSE attribute to use as a weight. Must be a number
@@ -783,23 +777,27 @@ class Rucio(object):
         Returns: A tuple of the chosen RSE and if the chosen RSE requires approval to write (rule property)
         """
         matchingRSEs = self.evaluateRSEExpression(rseExpression)
-        rsesWithWeights = []
+        rsesWithApproval = []
+        rsesWeight = []
 
         for rse in matchingRSEs:
-            attrs = self.cli.list_rse_attributes(rse)
+            rseAttrs = self.cli.list_rse_attributes(rse)
             if rseAttribute:
                 try:
-                    quota = float(attrs.get(rseAttribute, 0))
+                    attrValue = float(rseAttrs.get(rseAttribute, 0))
                 except (TypeError, KeyError):
-                    quota = 0
+                    attrValue = 0
             else:
-                quota = 1
-            requiresApproval = attrs.get('requires_approval', False)
-            if quota > minNeeded:
-                rsesWithWeights.append(((rse, requiresApproval), quota))
+                attrValue = 1
+            requiresApproval = rseAttrs.get('requires_approval', False)
+            if rseAttribute == "ddm_quota" and attrValue > minNeeded:
+                rsesWithApproval.append((rse, requiresApproval))
+                rsesWeight.append(attrValue)
+            elif rseAttribute != "ddm_quota":  # e.g. dm_weight
+                rsesWithApproval.append((rse, requiresApproval))
+                rsesWeight.append(attrValue)
 
-        choice = weightedChoice(rsesWithWeights)
-        return choice
+        return weightedChoice(rsesWithApproval, rsesWeight)
 
     def requiresApproval(self, rse):
         """
@@ -816,7 +814,6 @@ class Rucio(object):
             raise WMRucioException(msg)
         return attrs.get('requires_approval', False)
 
-
     def isContainer(self, didName, scope='cms'):
         """
         Checks whether the DID name corresponds to a container type or not.
@@ -924,7 +921,7 @@ class Rucio(object):
 
         # At this point, we might already have some of the RSEs where the data is available and locked
         # Now check dataset locks and compare those rules against our list of multi RSE rules
-        if self.isContainer(kwargs['name']):
+        if self.isContainer(kwargs['name'], scope=kwargs.get('scope', 'cms')):
             # It's a container! Find what those RSEs are and add them to the finalRSEs set
             rseLocks = self._getContainerLockedAndAvailable(multiRSERules, returnTape=returnTape, **kwargs)
             self.logger.debug("Data location for %s from multiple RSE locks and available at: %s",
@@ -983,17 +980,17 @@ class Rucio(object):
         rsesByBlocks = {}
         for block in blockNames:
             rsesByBlocks.setdefault(block, set())
-            ### FIXME: feature request made to the Rucio team to support bulk operations:
-            ### https://github.com/rucio/rucio/issues/3982
+            # FIXME: feature request made to the Rucio team to support bulk operations:
+            # https://github.com/rucio/rucio/issues/3982
             for blockLock in self.cli.get_dataset_locks(kwargs['scope'], block):
                 if not returnTape and isTapeRSE(blockLock['rse']):
                     continue
                 if blockLock['state'] == 'OK' and blockLock['rule_id'] in multiRSERules:
                     rsesByBlocks[block].add(blockLock['rse'])
 
-        ### The question now is:
-        ###   1. do we want to have a full copy of the container in the same RSEs (grouping=A)
-        ###   2. or we want all locations holding at least one block of the container (grouping=D)
+        # The question now is:
+        #   1. do we want to have a full copy of the container in the same RSEs (grouping=A)
+        #   2. or we want all locations holding at least one block of the container (grouping=D)
         if kwargs.get('grouping') == 'A':
             firstRun = True
             for _block, rses in viewitems(rsesByBlocks):
@@ -1031,7 +1028,7 @@ class Rucio(object):
         a specific method for this process, even though that adds some code duplication.
         """
         result = dict()
-        if not self.isContainer(container):
+        if not self.isContainer(container, scope=scope):
             raise WMRucioException("Pileup location needs to be resolved for a container DID type")
 
         multiRSERules = []
@@ -1139,7 +1136,7 @@ class Rucio(object):
         """
         if 'name' not in kwargs:
             raise WMRucioException("A DID name must be provided to the getBlockLockedAndAvailable API")
-        if self.isContainer(kwargs['name']):
+        if self.isContainer(kwargs['name'], scope=kwargs.get('scope', 'cms')):
             # then resolve it at container level and all its blocks
             return self.getContainerLockedAndAvailable(**kwargs)
 
@@ -1277,5 +1274,18 @@ class Rucio(object):
         else:
             finalRSEs = list(finalRSEs)
         self.logger.info("Container: %s with block-based location at: %s, and final location: %s",
-                          kwargs['name'], commonBlockRSEs, finalRSEs)
+                         kwargs['name'], commonBlockRSEs, finalRSEs)
         return finalRSEs
+
+    def getRSEUsage(self, rse):
+        """
+        get_rse_usage rucio API
+
+        :param rse: name of RSE
+        :return: generator of records about given RSE
+        """
+        try:
+            return self.cli.get_rse_usage(rse)
+        except Exception as exp:
+            self.logger.error("Failed to get information about rse %s. Error: %s", rse, str(exp))
+            raise exp
