@@ -200,10 +200,12 @@ class WorkQueue(object):
         if not elementIds:
             return
         eleParams = {}
+        if 'RequestPriority' in updatedParams:
+            updatedParams['Priority'] = updatedParams.pop('RequestPriority')
         eleParams[self.eleKey] = updatedParams
         conflictIDs = self.db.updateBulkDocumentsWithConflictHandle(elementIds, eleParams, maxConflictLimit=20)
         if conflictIDs:
-            raise CouchConflictError("WQ update failed with conflict", data=updatedParams, result=conflictIDs)
+            raise CouchConflictError("WQ update failed with conflict", data=updatedParams, result=conflictIDs, status=409)
         return
 
     def getAvailableWorkflows(self):
@@ -218,7 +220,18 @@ class WorkQueue(object):
         return availableSet - notAvailableSet
 
     def cancelWorkflow(self, wf):
-        """Cancel a workflow"""
+        """
+        Cancel a workflow in both workqueue and workqueue_inbox database.
+        :param wf: string with the workflow name
+        :return: the status code for the couchdb update call
+        """
+        # first, update the single workflow workqueue_inbox element, if needed
+        inboxData = self.getOpenRequests(inboxFlag=True, options={"key": wf})
+        if inboxData:
+            updateParams = {self.eleKey: {"OpenForNewData": False}}
+            self.inboxDB.updateBulkDocuments(inboxData, updateParams)
+
+        # now update all of the workqueue elements
         nonCancelableElements = ['Done', 'CancelRequested', 'Canceled', 'Failed']
         data = self.db.loadView('WorkQueue', 'elementsDetailByWorkflowAndStatus',
                                 {'startkey': [wf], 'endkey': [wf, {}],
@@ -239,12 +252,46 @@ class WorkQueue(object):
         # Update the spec, if it exists
         if self.db.documentExists(wf):
             wmspec = WMWorkloadHelper()
-            wmspec.load(self.db['host'] + "/%s/%s/spec" % (self.db.name, wf))
+            wmspec.load(self.hostWithAuth + "/%s/%s/spec" % (self.db.name, wf))
             wmspec.setPriority(priority)
             dummy_values = {'name': wmspec.name()}
             wmspec.saveCouch(self.hostWithAuth, self.db.name, dummy_values)
         return
 
+    def updateElementsByWorkflow(self, workload, updateParams, status=None):
+        """
+        Update all available WorkQueue elements of a given workflow  with a set
+        of arguments provided through the `updateParams` dictionary
+        :param workload:     A workflow spec
+        :param updateParams: A dictionary with parameters  to be updated
+        :param status:       A list of allowed WorkQueue elements statuses to be considered for updating
+                             Default: None - do not filter by status
+        :return:             No value, raises exceptions from internal methods in case of errors.
+        """
+        # Fetch the whole view with Workqueue elements per given workflow
+        wfName = workload.name()
+        data = self.db.loadView('WorkQueue', 'elementsDetailByWorkflowAndStatus',
+                                {'startkey': [wfName], 'endkey': [wfName, {}],
+                                 'reduce': False})
+
+        # Fetch only a list of WorkQueue element Ids && Filter them by allowed status
+        if status:
+            elementsToUpdate = [x['id'] for x in data.get('rows', []) if x['value']['Status'] in status]
+        else:
+            elementsToUpdate = [x['id'] for x in data.get('rows', [])]
+
+        # Update all WorkQueue elements with the parameters provided in a single push
+        if elementsToUpdate:
+            self.updateElements(*elementsToUpdate, **updateParams)
+
+        # Update the spec, if it exists
+        if self.db.documentExists(wfName):
+            # Update all workload parameters based on the full reqArgs dictionary
+            workload.updateWorkloadArgs(updateParams)
+            # Commit the changes of the current workload object to the database:
+            workload.saveCouchUrl(workload.specUrl())
+        return
+
     def getWorkflowNames(self, inboxFlag=False):
         """Get workflow names from workqueue db"""
         if inboxFlag:
@@ -296,7 +343,6 @@ class WorkQueue(object):
                                                 'Jobs': x['value']['sum']}
         return result
 
-
     def _retrieveWorkflowStatus(self, data):
         workflowsStatus = {}
 
@@ -307,7 +353,6 @@ class WorkQueue(object):
                 workflowsStatus[workflow] = status
         return workflowsStatus
 
-
     def getWorkflowStatusFromWQE(self, stale=True):
         """
         only checks workqueue db not inbox db.
@@ -403,3 +448,17 @@ class WorkQueue(object):
                 result[x['key']].append(x['doc'])
 
         return result
+
+    def getOpenRequests(self, inboxFlag=False, options=None):
+        """
+        Return workflow names that are still opened for new
+        data in workqueue
+        :param inboxFlag: boolean to select the *_inbox db
+        :param options: a dictionary that supports a few view parameters
+        :return: a list with unique request names
+        """
+        options = options or {}
+        options.setdefault("stale", "update_after")
+        db = self.inboxDB if inboxFlag else self.db
+        data = db.loadView('WorkQueue', 'openRequests', options)
+        return [x['key'] for x in data.get('rows', [])]
